-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util

local generator = { }
local data = { }
local tests = { }

-- generators and data do not exist yet
npcg.generators.alignment	= generator
npcg.data.alignment				= data
npcg.tests.alignment			= tests

-- Adding "any" and another will always return true.

local alignments = { }
alignments["lawful good"    ] = { moral = {    "good", "any" }, ethic = {  "lawful", "any" } }
alignments["lawful neutral" ] = { moral = { "neutral", "any" }, ethic = {  "lawful", "any" } }
alignments["lawful evil"    ] = { moral = {    "evil", "any" }, ethic = {  "lawful", "any" } }
alignments["neutral good"   ] = { moral = {    "good", "any" }, ethic = { "neutral", "any" } }
alignments["true neutral"   ] = { moral = { "neutral", "any" }, ethic = { "neutral", "any" } }
alignments["neutral evil"   ] = { moral = {    "evil", "any" }, ethic = { "neutral", "any" } }
alignments["chaotic good"   ] = { moral = {    "good", "any" }, ethic = { "chaotic", "any" } }
alignments["chaotic neutral"] = { moral = { "neutral", "any" }, ethic = { "chaotic", "any" } }
alignments["chaotic evil"   ] = { moral = {    "evil", "any" }, ethic = { "chaotic", "any" } }

local alignmentsByTag = { 
  ["vengeful"] = {
    "chaotic neutral",
    "lawful neutral",
    "lawful evil",
    "neutral evil",
    "chaotic evil" },
}

-- generateRandomAlignment :: CharacterTemplate -> Moral -> Alignment
function generateRandomAlignment(characterTemplate, moral)
	return util.randomFrom(
		functua.filterKeysByValue(
			function(val) return
				functua.contains(moral or "any",	val.moral) --[[or contains(params.moral, val.moral)]]
			end,
			alignments
		)
	) or "Error: No options found!"
end

-- alignmentsWithMoral :: Moral -> { Alignment }
function data.alignmentsWithMoral(moral)
  return functua.filterKeysByValue(
    function(val) return
      functua.contains(moral, val.moral)
    end,
    alignments
  )
end

function data.alignmentsByEthics(ethic)
  return functua.filterKeysByValue(
    function(val) return
      functua.contains(ethic, val.ethic)
    end,
    alignments
  )
end

-- allAlignmentsExcept :: Alignment -> ... -> { Alignment }
function data.allAlignmentsExcept(...)
  local args = {...}
  return functua.without(functua.keys(alignments), args)
end

function data.getAlignmentsByTag(tag)
  return alignmentsByTag[tag] or functua.keys(alignments)
end
-- End of File