-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util

local generator = { }
local data = { }
local tests = { }

-- generators and data do not exist yet
npcg.generators.class	= generator
npcg.data.class				= data
npcg.tests.class			= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Model | ---------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- default :: { TraitDepencency } -> { TraitDepencency }
-- Provides defaults for trait dependencies, so others don't have to fill them in.
local function default(replacements)
  local replacements = replacements or { }
  return {
    classTypes  = replacements["classTypes"] or {       },
    races       = replacements[     "races"] or { "any" },
    faiths      = replacements[    "faiths"] or { "any" },
    genders     = replacements[   "genders"] or { "any" },
    alignments  = replacements["alignments"] or { "any" },
    ages        = replacements[      "ages"] or npcg.data.age.olderThan("child"),
    weight      = 1
  }
end

-- The classes and their dependencies and classTypes regardless of locations or other context.
local classes = {
-- a
[ "adventurer"            ] = default({
  classTypes = {
    "explorer",
    "mercenary", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "alchemist"             ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "animalhandler"         ] = default({
  classTypes = {
    "animal",
    "hunter", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "antiquities dealer"    ] = default({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "archer"                ] = default({
  classTypes = {
    "military",
    "hunter",
    "mercenary",
    },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "archeologist"          ] = default({
  classTypes = {
    "scholar",
    "explorer", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "\"archmage\""          ] = default({
  classTypes = {
    "children" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf" },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "armorer"               ] = default({
  classTypes = {
    "crafter",
    "military", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "assassin"              ] = default({
  classTypes = {
    "rogue",
    "covert", },
  races = npcg.data.race.allRacesExcept(
  "draenei" ),
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "astrologer"            ] = default({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "astronomer"            ] = default({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "author"                ] = default({
  classTypes = {
    "scholar",
    "trader", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
-- b
[ "baker"                 ] = default({
  classTypes = {
    "food",
    "trader", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "bard"                  ] = default({
  classTypes = {
    "entertainer",
    "bar", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "bartender"             ] = default({
  classTypes = {
    "bar", },   
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "beggar"                ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "blacksmith"            ] = default({
  classTypes = {
    "crafter" },
  ages = npcg.data.age.agesBetween( "young adult", "adult" ) }),
[ "brat"                  ] = default({
  classTypes = {
    "children" },  
  ages = npcg.data.age.agesBetween( "child", "child" ) }),
[ "brawler"               ] = default({
  classTypes = {
    "bar",
    "mercenary", },
  races = npcg.data.race.allRacesExcept(
    "pandaren" ), 
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "blood mage"            ] = default({
  classTypes = {
    "mage",
    "scholar", },
  races = {
    "blood elf" },  
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "bodyguard"             ] = default({
  classTypes = {
    "mercenary", },   
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "bookbinder"            ] = default({
  classTypes = {
    "crafter",
    "scholar" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "undead",
    "worgen" },  
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "bounty hunter"         ] = default({
  classTypes = {
    "mercenary",
    "rogue", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "bowyer"                ] = default({
  classTypes = {
    "crafter",
    "hunter" },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "brewer"                ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "brothelkeeper"         ] = default({
  classTypes = {
    "illegal",
    "rogue" },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "orc",
    "pandaren",
    "troll",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan("young adult") }),
[ "brothelworker"         ] = default({
  classTypes = {
    "illegal",
    "rogue" },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "orc",
    "pandaren",
    "troll",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "burglar"               ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "teen", "adult" ) }),
[ "butcher"               ] = default({
  classTypes = {
    "food",
    "trader", },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- c
[ "candlemaker"           ] = default({
  classTypes = {
    "crafter", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "pandaren",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "carpenter"             ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "carpenter apprentice"  ] = default({
  classTypes = {
    "crafter",
    "apprentice" },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "clergy"                ] = default({
  classTypes = {
    "priest",
    "light", },
  faiths = npcg.data.faith.faithsByTag( "light" ),
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "clothier"              ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "cobbler"               ] = default({
  classTypes = {
    "crafter", },
  races = npcg.data.race.allRacesExcept(
    "draenei",
    "pandaren",
    "tauren" ),  
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "cobbler apprentice"    ] = default({
  classTypes = {
    "crafter",
    "apprentice" },
  ages = npcg.data.age.agesBetween( "child","teen" ) }),
[ "commoner"              ] = default({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "cook"                  ] = default({
  classTypes = {
    "food",
    "bar", },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "craftsman"             ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "cultist"               ] = default({
  classTypes = {
    "priest",
    "dark" },
  faiths = npcg.data.faith.faithsByTag( "cult" ), 
  ages = npcg.data.age.olderThan( "child" ) }),
-- d
[ "dark ranger"           ] = default({
  classTypes = {
    "hunter" },
  races = {
    "undead" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "demon hunter"          ] = default({
  classTypes = {
    "demon hunter" },
  races = {
    "blood elf", 
    "night elf" },
  faiths = {
    "the illidari" }, 
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "diplomat"              ] = default({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "doctor"                ] = default({
  classTypes = {
    "medical" },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "doomsayer"             ] = default({
  classTypes = {
    "town support",
    "dark" },
  faiths = npcg.data.faith.faithsByTag( "cult" ),
  ages = npcg.data.age.olderThan( "child" ) }),
[ "druid"                 ] = default({
  classTypes = {
    "druid" },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ) }),
[ "drummer"               ] = default({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "drunk"                 ] = default({
  classTypes = {
    "illegal",
    "town support", },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- e
[ "enchanter"             ] = default({
  classTypes = {
    "mage",
    "crafter" },  
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "engineer"              ] = default({
  classTypes = {
    "crafter",
    "inventor", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "entrepreneur"          ] = default({
  classTypes = {
    "trader" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "explorer"              ] = default({
  classTypes = {
    "explorer" },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- f
[ "farstrider"            ] = default({
  classTypes = {
    "military", "hunter" },
  races = {
    "blood elf" },
  ages = npcg.data.age.agesBetween( "adult", "mature adult" ) }),
[ "farmer"                ] = default({
  classTypes = {
    "gatherer" },
  ages = npcg.data.age.agesBetween( "young adult", "elder" ) }),
[ "fisherman"             ] = default({
  classTypes = {
    "gatherer",
    "harbor", },
  ages = npcg.data.age.agesBetween( "young adult", "adult" ) }),
[ "fishmonger"            ] = default({
  classTypes = {
    "trader" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "fruit vendor"          ] = default({
  classTypes = {
    "trader" },   
  ages = npcg.data.age.olderThan( "child" ) }),
-- g
[ "gang member"           ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "gardener"              ] = default({
  classTypes = {
    "town support" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "worgen" },
  ages = npcg.data.age.olderThan( "child") }),
[ "guard"                 ] = default({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "grunt"                 ] = default({
  classTypes = {
    "town support",
    "mercenary", },
  races = {
    "orc" },  
  ages = npcg.data.age.agesBetween( "young adult", "adult" ) }),
-- h
[ "hairdresser"           ] = default({
  classTypes = {
    "town support" },   
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "hay merchant"          ] = default({
  classTypes = {
    "trader" },   
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "headhunter"            ] = default({
  classTypes = {
    "mercenary", "hunter" },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "herald"                ] = default({
  classTypes = {
    "citizen" },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "orc",
    "pandaren",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "herbalist"             ] = default({
  classTypes = {
    "gatherer" },   
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "homeless person"       ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages =  npcg.data.age.agesBetween( "child", "child" ) }),
[ "horse trainer"         ] = default({
  classTypes = {
    "animal" },
  races = {
  "blood elf",
  "dwarf",
  "half elf",
  "half orc",
  "high elf",
  "human",
  "undead",
  "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "hunter"                ] = default({
  classTypes = {
    "hunter" },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
-- i
[ "illuminator"           ] = default({
  classTypes = {
    "town support", },
  races = {
    "blood elf",
    "dwarf",
    "goblin",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "orc",
    "pandaren",
    "tauren",
    "worgen",
    "undead" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "innkeeper"             ] = default({
  classTypes = {
    "bar",
    "citizen", },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "inventor"              ] = default({
  classTypes = {
    "inventor",
    "trader", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
-- j
[ "jailer"                ] = default({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "jester"                ] = default({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "jewelcrafter"          ] = default({
  classTypes = {
    "crafter",
    "trader" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "juggler"               ] = default({
  classTypes = {
    "entertainer" },   
  ages = npcg.data.age.agesBetween("child", "elder") }),
-- k
[ "kitchen help"          ] = default({
  classTypes = {
    "bar",
    "food", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "knight"                ] = default({
  classTypes = {
    "paladin",
    "military" },
  races = {
    "human" },  
  ages = npcg.data.age.olderThan( "adult" ) }),
-- l
[ "laborer"               ] = default({
  classTypes = {
    "citizen" },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ) }),
[ "leatherworker"         ] = default({
  classTypes = {
    "crafter", "hunter" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "loan provider"         ] = default({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "lookout"               ] = default({
  classTypes = {
    "rogue",
    "mercenary" },
  ages = npcg.data.age.agesBetween( "child", "young adult" ) }),
[ "lorewalker"            ] = default({
  classTypes = {
    "scholar",
    "explorer", },
  races = {
    "pandaren" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "lumberjack"            ] = default({
  classTypes = {
    "gatherer" },   
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ) }),
-- m
[ "mage"                  ] = default({
  classTypes = {
    "mage",
    "scholar", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ) }),
[ "mathmatician"          ] = default({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "high elf",
    "human",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "medic"                 ] = default({
  classTypes = {
    "medical" },
  ages = npcg.data.age.agesBetween( "young adult", "elder" ) }),
[ "mercenary"             ] = default({
  classTypes = {
    "mercenary",
    "warrior", },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ) }),
[ "merchant"              ] = default({
  classTypes = {
    "trader" },
  ages = npcg.data.age.agesBetween( "teen", "elder" ) }),
[ "messenger"             ] = default({
  classTypes = {
    "citizen" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "military advisor"      ] = default({
  classTypes = {
    "military", },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "mime"                  ] = default({
  classTypes = {
    "entertainer" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "miner"                 ] = default({
  classTypes = {
    "gatherer" },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ) }),
[ "monk"                  ] = default({
  classTypes = {
    "monk" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "musician"              ] = default({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- n
[ "nanny"                 ] = default({
  classTypes = {
    "children",
    "citizen", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "navigator"             ] = default({
  classTypes = {
    "harbor", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "necromancer"           ] = default({
  classTypes = {
    "cult", },
  faiths = {
    "cult of the damned" },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "nurse"                 ] = default({
  classTypes = {
    "medical", },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- o
[ "old clothes dealer"    ] = default({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "outlaw"                ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" ) }),
-- p
[ "page"                  ] = default({
  classTypes = {
    "paladin", 
    "apprentice", },
  races = {
    "blood elf",
    "dwarf",
    "draenei",
    "goblin",
    "half elf",
    "high elf",
    "human",
    "night elf",
    "tauren" },
  ages = npcg.data.age.agesBetween( "child", "child" ) }),
[ "painter"               ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "paladin"               ] = default({
  classTypes = {
    "paladin", },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "tauren" },
  faiths = npcg.data.faith.faithsByTag( "light" ),
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "peasant"               ] = default({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "pharmacist"            ] = default({
  classTypes = {
    "trader", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "philosopher"           ] = default({
  classTypes = {
    "scholar", },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "physician"             ] = default({
  classTypes = {
    "medical", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "pickpocket"            ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "pie seller"            ] = default({
  classTypes = {
    "trader", },   
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "piper"                 ] = default({
  classTypes = {
    "crafter", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "tauren",
    "troll",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "piper apprentice"      ] = default({
  classTypes = {
    "crafter",
    "apprentice" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "tauren",
    "troll",
    "worgen" },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "pirate"                ] = default({
  classTypes = {
    "harbor",
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "potter"                ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "potter apprentice"     ] = default({
  classTypes = {
    "crafter", 
    "apprentice", },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "preacher"              ] = default({
  classTypes = {
    "priest", 
    "light", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "priest"                ] = default({
  classTypes = {
    "priest",
    "light", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ) }),
-- q
-- r
[ "ranger"                ] = default({
  classTypes = {
    "hunter", },
  races = {
    "blood elf",
    "human",
    "high elf",
    "undead" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "rascal"                ] = default({
  classTypes = {
    "children", },
  ages = npcg.data.age.agesBetween( "child", "child" ) }),
[ "rebel"                 ] = default({
  classTypes = {
    "mercenary", },   
  ages = npcg.data.age.olderThan( "child" ) }),
-- s
[ "sailor"                ] = default({
  classTypes = {
    "harbor", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "scholar"               ] = default({
  classTypes = {
    "scholar", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "scout"                 ] = default({
  classTypes = {
    "hunter",
    "mercenary",
    "military", },   
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "scribe"                ] = default({
  classTypes = {
    "scholar", },
  races = npcg.data.race.racesByType( "civilized" ),  
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "scullion"              ] = default({
  classTypes = {
    "crafter", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "servant"               ] = default({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "sewer"                 ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "shadow hunter"         ] = default({
  classTypes = {
    "dark" },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "shaman"                ] = default({
  classTypes = {
    "shaman", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ) }),
[ "shepherd"              ] = default({
  classTypes = {
    "farmer" },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "shopkeeper"            ] = default({
  classTypes = {
    "trader", },
  ages = npcg.data.age.agesBetween( "young adult" , "mature adult" ) }),
[ "skinner"               ] = default({
  classTypes = {
    "gatherer",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "skinner apprentice"    ] = default({
  classTypes = {
    "apprentice",
    "hunter",
    "trader", },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "soldier"               ] = default({
  classTypes = {
    "military",
    "hunter",
    "paladin", },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ) }),
[ "sorcerer"              ] = default({
  classTypes = {
    "mage",
    "warlock", },   
  ages = npcg.data.age.olderThan("young adult") }),
[ "spellbreaker"          ] = default({
  classTypes = {
    "town support",
    "military", },
  races = {
    "blood elf" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "spy"                   ] = default({
  classTypes = {
    "rogue",
    "covert", },
  races = npcg.data.race.allRacesExcept(
    "draenei",
    "tauren" ),
  ages = npcg.data.age.olderThan("child") }),
[ "squire"                ] = default({
  classTypes = {
    "apprentice", 
    "paladin", },
  races = {
    "blood elf",
    "draenei",
    "goblin",
    "half elf",
    "high elf",
    "human",
    "night elf",
    "tauren" },
  ages = npcg.data.age.agesBetween( "child", "young adult" ) }),
[ "stablekeeper"          ] = default({
  classTypes = {
    "animal",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "stablehand"            ] = default({
  classTypes = {
    "animal",
    "hunter" },
  ages = npcg.data.age.agesBetween( "child", "young adult" ) }),
[ "student"               ] = default({
  classTypes = {
    "children",
    "scholar",
    "priest",
    "mage" },
  ages = npcg.data.age.agesBetween( "child", "teen" ) }),
[ "surgeon"               ] = default({
  classTypes = {
    "medical" },
  races = npcg.data.race.racesByType( "civilized" ), 
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "swindler"              ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" ) }),
-- t
[ "tailor"                ] = default({
  classTypes = {
    "crafter",
    "trader", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "teacher"               ] = default({
  classTypes = {
    "children",
    "scholar", },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "thief"                 ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "child", "mature adult" ) }),
[ "thug"                  ] = default({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "tinker"                ] = default({
  classTypes = {
    "crafter",
    "inventor", },
  races = {
    "gnome",
    "goblin" },
  ages = npcg.data.age.olderThan( "young adult" ) }),
[ "trapper"               ] = default({
  classTypes = {
    "gatherer",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- u 
-- v
[ "vendor"                ] = default({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- w
[ "waiter"                ] = default({
  classTypes = {
    "bar", },
  ages = npcg.data.age.agesBetween( "child", "mature adult" ) }),
[ "warlock"               ] = default({
  classTypes = {
    "warlock",
    "dark" },   
  ages = npcg.data.age.agesBetween( "young adult",  "venerable" ) }),
[ "watchman"              ] = default({
  classTypes = {
    "town support",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "water vendor"          ] = default({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "wine vendor"           ] = default({
  classTypes = {
    "trader", },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "undead" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "witch doctor"          ] = default({
  classTypes = {
    "nature" },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "adult" ) }),
[ "woodcarver"            ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" ) }),
[ "woodworker"            ] = default({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ) }),
[ "woolseller"            ] = default({
  classTypes = {
    "trader", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half elf",
    "high elf",
    "human",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
[ "wool-stapler"          ] = default({
  classTypes = {
    "crafter", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "undead",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" ) }),
-- x
-- y
-- z
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Queries | ------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- classesByType :: ClassType -> { Class }
function data.classesByType(classType)
  return functua.filterKeysByValue(
    function(x) return
      functua.contains(classType, x.classTypes)
    end,
    util.eitherOr(classType == "any", { }, classes)
  )
end

-- classesByType2 :: ClassType -> { Class, ClassConstraints }
function data.classesByType2(classType)
  -- this should be horrible?
  local classesWithClassType = data.classesByType(classType)
  local output = { }
  for _,v in ipairs(classesWithClassType) do -- Or is it not..? I can't wrap my head around it >.<
    output[v] = classes[v]
  end
  return output
end

-- This takes an Any approach
function data.classesByAnyOfTypes(...)
  return functua.mergeTablesWithoutDuplicates(unpack(funtua.map(classesByType, {...})))
end

function data.classesByAnyOfTypes2(...)
  return functua.mergeTablesWithoutDuplicates2(unpack(functua.map(classesByType2, {...})))
end

-- allClassesExcept :: Class -> ... -> { Class }
function data.allClassesExcept(...)
  local exceptions = {...}
  return functua.without(functua.keys(classes),exceptions)
end

-- randomClass :: () -> Class
-- Generates a true random class without constraints.
-- This is a fallback that should not be used all too lightly.
function data.randomClass()
  return util.randomFrom(functua.keys(classes))
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Location Based Generation | ------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------
local zonesToClasses = {
  ["Dalaran"              ] = nil,
  -- Order Halls
  ["Dreadscar Rift"       ] = data.classesByType2("warlock"),
  ["Hall of the Guardian" ] = data.classesByType2("mage"),
  ["Netherlight Temple"   ] = data.classesByType2("priest"),
  ["Skyhold"              ] = data.classesByType2("warrior"),
  ["The Dreamgrove"       ] = data.classesByType2("druid"),
  ["Trueshot Lodge"       ] = data.classesByType2("hunter"),
  
  -- Starter Zones
  ["Eversong Woods"       ] = nil,
}

local subzonesToClasses = {
  [ "Silvermoon City" ] = {
    ["Murder Row"         ] = data.classesByType2("rogue")
  },
  [ "Dalaran"         ] = {
    -- Rogue Order Hall
    ["Chamber of Shadows" ] = data.classesByType2("rogue"),
    ["Cloaked Vestibule"  ] = data.classesByType2("rogue"),
    ["Hidden Corridor"    ] = data.classesByType2("rogue"),
    ["The Hall of Shadows"] = data.classesByType2("rogue"),
    ["Uncrowned Vault"    ] = data.classesByType2("rogue"),
  },
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local function checkAllLocations(characterTemplate, zone, subZone)
  return 
    -- Test whether the SubZone has values.
    ((subzonesToClasses[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToClasses[zone]
    -- Fallback to default.
    or classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function checkZonesOnly(characterTemplate, zone)
  return
    -- Test whether the Zone has values.
    zonesToClasses[zone]
    -- Fallback to default.
    or classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function dontCheckLocations(characterTemplate)
  return
    -- Fallback to default.
    classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- characterTemplate -> class -> zone -> subzone -> bool

local function meetsTraitConstraints(characterTemplate, option)
  return  functua.containsAny(classes[option].races,   { "any", characterTemplate.race   })
    and   functua.containsAny(classes[option].ages,    { "any", characterTemplate.age    })
    and   functua.containsAny(classes[option].faiths,  { "any", characterTemplate.faith  })
end

local function isWowClass(characterTemplate, option)
  return functua.any(
    functua.composeEqualsFunction(option),
    { "druid", "hunter", "mage", "monk", "paladin", "priest", "rogue", "shaman", "warlock", "warrior" })
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- Arbitrary a => { class, a } -> class
-- This might change in the future.

local function randomKeyFrom(options)
  return util.randomFrom(functua.keys(options))
end

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

function createClassGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone)
    return selectorF(
      functua.filter2(
        function(class) return filterF(characterTemplate, class, zone, subZone) end, 
        generatorF(characterTemplate, zone, subZone)
      )
    )
  end
end

local traitAndSubZoneBasedClassGenerator          = createClassGenerator(checkAllLocations,   meetsTraitConstraints,  selectUsingWeightedRandom)
local traitAndZoneBasedClassGenerator             = createClassGenerator(checkZonesOnly,      meetsTraitConstraints,  selectUsingWeightedRandom)
local traitAndNoLocationBasedClassGenerator       = createClassGenerator(dontCheckLocations,  meetsTraitConstraints,  selectUsingWeightedRandom)
local wowOnlyAndSubZoneBasedClassGenerator        = createClassGenerator(checkAllLocations,   isWowClass,             selectUsingWeightedRandom)
local wowOnlyAndZoneBasedClassGenerator           = createClassGenerator(checkZonesOnly,      isWowClass,             selectUsingWeightedRandom)
local wowOnlyAndNoLocationBasedClassGenerator     = createClassGenerator(dontCheckLocations,  isWowClass,             selectUsingWeightedRandom)
local noContextAndSubZoneBasedClassGenerator      = createClassGenerator(checkAllLocations,   functua.const(true),    selectUsingWeightedRandom)
local noContextAndZoneBasedClassGenerator         = createClassGenerator(checkZonesOnly,      functua.const(true),    selectUsingWeightedRandom)
local noContextAndNoLocationBasedClassGenerator   = createClassGenerator(dontCheckLocations,  functua.const(true),    selectUsingWeightedRandom)

function generator.initialize()
  generateRandomClass = function() return "Error: generateRandomClass is not set" end
  
  if not NPCG_Options.locationSettings.zones then
    generateRandomClass = function(characterTemplate)
      return traitAndNoLocationBasedClassGenerator(characterTemplate)
      or noContextAndNoLocationBasedClassGenerator(characterTemplate) end
  elseif NPCG_Options.locationSettings.subZones then
    generateRandomClass = function(characterTemplate, zone, subZone)
      return traitAndSubZoneBasedClassGenerator(characterTemplate, zone, subZone)
      or noContextAndSubZoneBasedClassGenerator(characterTemplate, zone, subZone)
    end
  else
    generateRandomClass = function(characterTemplate, zone)
      return traitAndZoneBasedClassGenerator(characterTemplate, zone)
      or noContextAndZoneBasedClassGenerator(characterTemplate, zone)
    end
  end
end

--[[ LEGACY: generateRandomClass
-- generateRandomClass :: (Race a, Age a) => a -> Class
function generateRandomClass(characterTemplate)
  local selectionList = 
    -- Subzone-based selection.
    ((subzonesToClasses[GetZoneText()] or { })[GetSubZoneText()])
    -- Zone-based selection.
    or zonesToClasses[GetZoneText()]
    -- Default
    or classes
  
  return randomFrom(
    filterKeysByValue(
      function(val) return
        (
          contains(characterTemplate:getRace(),   val.races) or 
          contains("any", val.races)
        ) and (
          contains(characterTemplate:getAge(),     val.ages) or
          contains("any", val.ages)
        ) and (
          contains(characterTemplate:getFaith(), val.faiths) or
          contains("any", val.faiths)
        )
      end,
      selectionList
    )
  ) or randomClass()
end
]]--

function tests.testClassesByType2()
  print "Testing classesByType2(\"rogue\")"
  local results = data.classesByType2("rogue")
  print "Keys"
  util.logTable(functua.keys(results))
  print "Values"
  functua.map(util.logTableOfPropertyLists, results)
  print "Ended testing"
end

-- End of File.